'use strict';

const { expect } = require('chai');
const mock = require('../../tools/mongodb-mock/index');
const { Topology } = require('../../../src/sdam/topology');
const { MongoCredentials } = require('../../../src/cmap/auth/mongo_credentials');
const { isHello } = require('../../../src/utils');

describe('SCRAM Iterations Tests', function () {
  let server;

  beforeEach(() => {
    return mock.createServer().then(mockServer => {
      server = mockServer;
    });
  });

  afterEach(() => mock.cleanup());

  it('should error if iteration count is less than 4096', function (_done) {
    const scramResponse =
      'r=IE+xNFeOcslsupAA+zkDVzHd5HfwoRuP7Wi8S4py+erf8PcNm7XIdXQyT52Nj3+M,s=AzomrlMs99A7oFxDLpgFvVb+CSvdyXuNagoWVw==,i=4000';

    const credentials = new MongoCredentials({
      mechanism: 'DEFAULT',
      source: 'db',
      username: 'user',
      password: 'pencil'
    });

    let done = e => {
      done = () => {};
      return _done(e);
    };

    server.setMessageHandler(request => {
      const doc = request.document;
      if (isHello(doc)) {
        return request.reply(Object.assign({}, mock.HELLO));
      } else if (doc.saslStart) {
        return request.reply({
          ok: 1,
          done: false,
          payload: Buffer.from(scramResponse)
        });
      } else if (doc.saslContinue) {
        done('SHOULD NOT BE HERE');
      }
    });

    const client = new Topology(server.hostAddress(), { credentials });
    client.on('error', err => {
      let testErr;
      try {
        expect(err).to.not.be.null;
        expect(err)
          .to.have.property('message')
          .that.matches(/Server returned an invalid iteration count/);
      } catch (e) {
        testErr = e;
      }
      client.close();
      done(testErr);
    });

    client.connect();
  });

  it('should error if server digest is invalid', function (_done) {
    const credentials = new MongoCredentials({
      mechanism: 'DEFAULT',
      source: 'db',
      username: 'user',
      password: 'pencil'
    });

    let done = e => {
      done = () => {};
      return _done(e);
    };

    server.setMessageHandler(request => {
      const doc = request.document;
      if (isHello(doc)) {
        return request.reply(Object.assign({}, mock.HELLO));
      } else if (doc.saslStart) {
        return request.reply({
          ok: 1,
          done: false,
          payload: Buffer.from(
            'r=VNnXkRqKflB5+rmfnFiisCWzgDLzez02iRpbvE5mQjMvizb+VkSPRZZ/pDmFzLxq,s=dZTyOb+KZqoeTFdsULiqow==,i=10000'
          )
        });
      } else if (doc.saslContinue) {
        return request.reply({
          ok: 1,
          done: false,
          payload: Buffer.from('v=bWFsaWNpb3VzbWFsaWNpb3VzVzV')
        });
      }
    });

    const client = new Topology(server.hostAddress(), { credentials });
    client.on('error', err => {
      expect(err).to.not.be.null;
      expect(err)
        .to.have.property('message')
        .that.matches(/Server returned an invalid signature/);

      client.close(done);
    });

    client.connect();
  });

  it('should properly handle network errors on `saslContinue`', function (_done) {
    const credentials = new MongoCredentials({
      mechanism: 'DEFAULT',
      source: 'db',
      username: 'user',
      password: 'pencil'
    });

    let done = e => {
      done = () => {};
      return _done(e);
    };

    server.setMessageHandler(request => {
      const doc = request.document;
      if (isHello(doc)) {
        return request.reply(Object.assign({}, mock.HELLO));
      } else if (doc.saslStart) {
        return request.reply({
          ok: 1,
          done: false,
          payload: Buffer.from(
            'r=VNnXkRqKflB5+rmfnFiisCWzgDLzez02iRpbvE5mQjMvizb+VkSPRZZ/pDmFzLxq,s=dZTyOb+KZqoeTFdsULiqow==,i=10000'
          )
        });
      } else if (doc.saslContinue) {
        request.connection.destroy();
      }
    });

    const client = new Topology(server.hostAddress(), { credentials });
    client.on('error', err => {
      expect(err).to.not.be.null;
      expect(err)
        .to.have.property('message')
        .that.matches(/connection(.+)closed/);

      client.close(done);
    });

    client.connect();
  });
});
